
#define ALG "E2"		/*** Customize - algorithm name ***/
#include "o-e2.h"


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>


#define	CPUID DW 0a20fh
#define	RDTSC DW 0310fh

#define MAX_LOOPS	1000

unsigned int	TimeEKey(int keysize);
unsigned int	TimeDKey(int keysize);
unsigned int	TimeEncrypt(int keysize);
unsigned int	TimeDecrypt(int keysize);
unsigned int	TimeCipherInit(int keysize);

void	processData(void);
double	StDev(int ind, int med);
int	Median(int ind);
unsigned long	AvgIf(int ind, double dev, int med);
void	outputRawData(int keysize);

int	data[MAX_LOOPS][5];
int	num_loops;

FILE	*raw;
FILE	*time_fp;

int main(int argc, char *argv[])
{
	int			i, keysize;
	char			filename[32];
	unsigned int	base=0;
	unsigned int	cyc, cycles[10];

	if ( argc != 3 ) {
		printf("usage: %s <keysize> <number of loops>\n", strrchr(argv[0], '\\')+1);
		exit(-1);
		}
	sprintf(filename, "..\\Adata\\%s.raw", ALG);
	if ( (raw = fopen(filename, "w")) == NULL ) {
		fprintf(stderr, "couldn't open <%s> for write\n", filename);
		exit(-1);
		}
	if ( (time_fp = fopen("..\\Adata\\cycles.tim", "a")) == NULL ) {
		fprintf(stderr, "couldn't open <..\\Adata\\cycles.tim> for append\n");
		exit(-1);
		}
	keysize = atoi(argv[1]);
	num_loops = atoi(argv[2]);
	if ( num_loops > MAX_LOOPS )
		num_loops = MAX_LOOPS;

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cyc, eax
		popad
		pushad
		CPUID
		RDTSC
		sub	eax, cyc
		mov	base, eax
		popad

		pushad
		CPUID
		RDTSC
		mov	cyc, eax
		popad
		pushad
		CPUID
		RDTSC
		sub	eax, cyc
		mov	base, eax
		popad

		pushad
		CPUID
		RDTSC
		mov	cyc, eax
		popad
		pushad
		CPUID
		RDTSC
		sub	eax, cyc
		mov	base, eax
		popad
		}

	fprintf(time_fp, "%s, %d", ALG, keysize);
	for ( i=0; i<num_loops; i++ )
		data[i][0] = TimeEKey(keysize)-base;
	for ( i=0; i<num_loops; i++ )
		data[i][1] = TimeDKey(keysize)-base;
	for ( i=0; i<num_loops; i++ )
		data[i][2] = TimeEncrypt(keysize)-base;
	for ( i=0; i<num_loops; i++ )
		data[i][3] = TimeDecrypt(keysize)-base;
	for ( i=0; i<num_loops; i++ )
		data[i][4] = TimeCipherInit(keysize)-base;

	outputRawData(keysize);
	fclose(raw);

	processData();
	}

unsigned int	TimeEKey(int keysize)
{
	unsigned int	cycles;

	char	*key_mat = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F";
	char	key_temp[65];
	keyInstance		key;		/* AES keyInstances */

	memcpy(key_temp, key_mat, (keysize+3)/4);
	key_temp[(keysize+3)/4] = '\0';

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cycles, eax
		popad
		}
	
	makeKey(&key, DIR_ENCRYPT, keysize, key_temp);

	__asm {
		pushad
		CPUID
		RDTSC
		sub	eax, cycles
		mov	cycles, eax
		popad
		}

	return cycles;
	}

unsigned int	TimeDKey(int keysize)
{
	unsigned int	cycles;

	char	*key_mat = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F";
	char	key_temp[65];
	keyInstance		key;		/* AES keyInstances */

	memcpy(key_temp, key_mat, (keysize+3)/4);
	key_temp[(keysize+3)/4] = '\0';

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cycles, eax
		popad
		}
	
	makeKey(&key, DIR_DECRYPT, keysize, key_temp);

	__asm {
		pushad
		CPUID
		RDTSC
		sub	eax, cycles
		mov	cycles, eax
		popad
		}

	return cycles;
	}

unsigned int	TimeEncrypt(int keysize)
{
	unsigned int	cycles;

	char	*key_mat = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F";
	char	key_temp[65];
	keyInstance		key;		/* AES keyInstances */
	cipherInstance	cipher;
	BYTE	pt[16], ct[16];
	char	iv[] = "00000000000000000000000000000000";
	int	ret_val;

	memcpy(key_temp, key_mat, (keysize+3)/4);
	key_temp[(keysize+3)/4] = '\0';
	ret_val = makeKey(&key, DIR_ENCRYPT, keysize, key_temp);
	if ( (ret_val != TRUE) ) {
		fprintf(stderr, "makeKey: ret_val is <%d>\n", ret_val);
		exit(-1);
		}
	ret_val = cipherInit(&cipher, MODE_ECB, iv);
	if ( (ret_val != TRUE) ) {
		fprintf(stderr, "cipherInit: ret_val is <%d>\n", ret_val);
		exit(-1);
		}

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cycles, eax
		popad
		}
	
	ret_val = blockEncrypt(&cipher, &key, pt, sizeof(pt)*8, ct);

	__asm {
		pushad
		CPUID
		RDTSC
		sub	eax, cycles
		mov	cycles, eax
		popad
		}

	return cycles;
	}

unsigned int	TimeDecrypt(int keysize)
{
	unsigned int	cycles;

	char	*key_mat = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F";
	char	key_temp[65];
	keyInstance		key;		/* AES keyInstances */
	cipherInstance	cipher;
	BYTE	pt[16], ct[16];
	char	iv[] = "00000000000000000000000000000000";
	int	ret_val;

	memcpy(key_temp, key_mat, (keysize+3)/4);
	key_temp[(keysize+3)/4] = '\0';
	ret_val = makeKey(&key, DIR_DECRYPT, keysize, key_temp);
	if ( (ret_val != TRUE) ) {
		fprintf(stderr, "makeKey: ret_val is <%d>\n", ret_val);
		exit(-1);
		}
	ret_val = cipherInit(&cipher, MODE_ECB, iv);
	if ( (ret_val != TRUE) ) {
		fprintf(stderr, "cipherInit: ret_val is <%d>\n", ret_val);
		exit(-1);
		}

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cycles, eax
		popad
		}
	
	ret_val = blockDecrypt(&cipher, &key, pt, sizeof(pt)*8, ct);

	__asm {
		pushad
		CPUID
		RDTSC
		sub	eax, cycles
		mov	cycles, eax
		popad
		}

	return cycles;
	}

unsigned int	TimeCipherInit(int keysize)
{
	unsigned int	cycles;

	cipherInstance	cipher;
	char	iv[] = "00000000000000000000000000000000";
	int	ret_val;

	__asm {
		pushad
		CPUID
		RDTSC
		mov	cycles, eax
		popad
		}
	
	cipherInit(&cipher, MODE_ECB, iv);

	__asm {
		pushad
		CPUID
		RDTSC
		sub	eax, cycles
		mov	cycles, eax
		popad
		}

	return cycles;
	}

void	processData(void)
{
	int	ind;
	double	dev, factor, secs;
	int	med;
	unsigned long	avg;

	for ( ind=0; ind<5; ind++ ) {
		med = Median(ind);
		dev = StDev(ind, med);
		avg = AvgIf(ind, dev*3, med);
		fprintf(time_fp, ", %d", avg);
		}
	fprintf(time_fp, "\n");
	}

double	StDev(int ind, int med)
{
	double	sum = 0.0, dev, scale;
	int	i, count=0;

	scale = (1.0 * med)/10000.0;
	for ( i=0; i<num_loops; i++ ) {
		/*  throw out outliers  */
		if ( data[i][ind] > 2*med )
			break;
		sum += (data[i][ind]/scale - med/scale)*(data[i][ind]/scale - med/scale);
		count++;
		}
	dev = sqrt(sum/(count-1));
	dev = dev * scale;
	return (dev);
	}

int	Median(int ind)
{
	int	i, j, temp;

	for ( i=0; i<num_loops/2+1; i++ )
		for ( j=i; j<num_loops; j++ )
			if ( data[i][ind] > data[j][ind] ) {
				temp = data[j][ind];
				data[j][ind] = data[i][ind];
				data[i][ind] = temp;
				}
	return (data[i-1][ind]);
	}

unsigned long	AvgIf(int ind, double dev, int med)
{
	int	i, count = 0;
	unsigned long	total = 0;

	for ( i=0; i<num_loops; i++ )
		if ( abs(data[i][ind] - med) <= dev ) {
			total += data[i][ind];
			count++;
			}
	if ( count == 0 )
		return (0);
	else
		return (total/count);
	}

void	outputRawData(int keysize)
{
	int	i;

	for ( i=0; i<num_loops; i++ )
		fprintf(raw, "%s, %d, %d, %d, %d, %d, %d\n", ALG, keysize,
			data[i][0], data[i][1], data[i][2], data[i][3], data[i][4]);
	}